//---------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <vcl.h>
#pragma hdrstop

//---------------------------------------------------------------------------
typedef unsigned char uchar;

#define usage() printf("usage: img2vhd_generator.exe source_path.bmp [-rgb [bits]]\n\n")

// ====== log2() ======
int log2(int x)
{
	int i;
	for(i=0;i<30;i++)
		if((1<<i)>x)
			break;
	return(i);
}

#pragma argsused
int main(int argc, char* argv[])
{
	// header
	printf("BMP to VHDL converter by Stanislav Maslan, s.maslan@seznam.cz, V1.10, 2012.05.13\n");

	// check parameters
	if(argc<2)
	{
		usage();
		printf("Error - missing parameter!\n\n");
		return(1);
	}

	// rgb mode enabled?
	int rgb=0;
	if(argc>2)
		rgb=strcmpi(argv[2],"-rgb")==0;

	// limit bits
	int bits=8;
	if(argc>3)
		bits=atoi(argv[3]);
	if(bits<1)
		bits=1;
	else if(bits>8)
		bits=8;

	// try to create bitmap
	Graphics::TBitmap *bmp;
	try{
		bmp=new Graphics::TBitmap;
	}catch(Exception &exc){
		// failed
		usage();
		printf("Error - creating TBitmap failed with exception:\n\"%s\"\n\n",exc.Message.c_str());
		return(1);
	}

	// try to open source bitmap
	try{
		bmp->LoadFromFile(AnsiString(argv[1]));
	}catch(Exception &exc){
		// failed
		usage();
		printf("Error - opening image failed with exception:\n\"%s\"\n\n",exc.Message.c_str());
		return(1);
	}

	// check its format
	if(bmp->PixelFormat!=pf24bit)
	{
		// invalid format
		usage();
		printf("Error - invalid format, image must be 24-bit raster!\n\n");
		delete(bmp);
		return(1);
	}

	// get bitmap resolution
	int x=bmp->Width;
	int y=bmp->Height;

	// loose path extentsion
	char path[260];
	strcpy(path,argv[1]);
	char *pstr=strrchr(path,'.');
	if(pstr)
		*pstr='\0';

	// loose folders if some
	char c='\0';
	while(pstr>path && (c=*--pstr)!='\\' && c!='/')
		;
	if(c=='\\' || c=='/')
		pstr++;
	char fname[260];
	// store pure file name
	strcpy(fname,pstr);

	// generate destination path
	strcat(path,".vhd");

	// try to create destination file
	FILE *fw=fopen(path,"wt");
	if(!fw)
	{
		usage();
		printf("Error - creating destination file failed!\n\n");
		delete(bmp);
		return(1);
	}

	// print some file headers
	fprintf(fw,"-- BITMAP GENERATOR --\n");
	fprintf(fw,"-- Source bitmap: %s\n\n\n",argv[1]);

	fprintf(fw,"library ieee;\n");
	fprintf(fw,"use ieee.std_logic_1164.all;\n");
	fprintf(fw,"use ieee.std_logic_arith.all;\n");
	fprintf(fw,"use ieee.std_logic_unsigned.all;\n\n");

	fprintf(fw,"package %s is\n",fname);
	fprintf(fw,"\t--- bitmap pixels ---\n");
	if(rgb)
	{
		fprintf(fw,"\tfunction pix_r(col: std_logic_vector; row: std_logic_vector) return std_logic_vector;\n");
		fprintf(fw,"\tfunction pix_g(col: std_logic_vector; row: std_logic_vector) return std_logic_vector;\n");
		fprintf(fw,"\tfunction pix_b(col: std_logic_vector; row: std_logic_vector) return std_logic_vector;\n\n");
	}
	else
		fprintf(fw,"\tfunction pixels (col: std_logic_vector; row: std_logic_vector) return std_logic;\n\n");
	fprintf(fw,"\t--- bitmap size ---\n");
	fprintf(fw,"\tconstant bmpx: integer := %d;\n",x);
	fprintf(fw,"\tconstant bmpy: integer := %d;\n",y);
	fprintf(fw,"\tconstant bitx: integer := %d;\n",log2(x-1));
	fprintf(fw,"\tconstant bity: integer := %d;\n",log2(y-1));
	fprintf(fw,"end %s;\n\n",fname);

	fprintf(fw,"package body %s is\n",fname);
	if(rgb)
	{
		uchar msk[9]={0x00,0x80,0xC0,0xE0,0xF0,0xF8,0xFC,0xFE,0xFF};
		char *fun[3]={"pix_r","pix_g","pix_b"};
		char *name[3]={"red","green","blue"};
		int bitw[3]={5,6,5};

		int k;
		for(k=0;k<3;k++)
		{
			// color function
			fprintf(fw,"\t--- bitmap %s pixels ---\n",name[k]);
			fprintf(fw,"\tfunction %s(col: std_logic_vector; row: std_logic_vector) return std_logic_vector is\n",fun[k]);
			fprintf(fw,"\t\tvariable pix: integer range 0 to 2**%d-1;\n",bitw[k]);
			fprintf(fw,"\tbegin\n");

			// convert color layer
			int i,j;
			for(j=0;j<y;j++)
			{
				uchar *pix=(uchar*)(bmp->ScanLine[j]);
				for(i=0;i<x;i++)
				{
					uchar px=(pix[2-k]&msk[bits])>>(8-bitw[k]);

					fprintf(fw,"\t\t%s (col=%-3d and row=%-3d) then pix := %d;\n",(i!=0 || j!=0)?"elsif":"   if",i,j,px);
					pix+=3;
				}
			}

			// generate rest of VHDL trash
			fprintf(fw,"\t\telse pix := %d; end if;\n\n");
			fprintf(fw,"\t\treturn conv_std_logic_vector(pix,%d);\n",bitw[k]);
			fprintf(fw,"\tend %s;\n\n",fun[k]);

		}
	}
	else
	{
		fprintf(fw,"\t--- bitmap pixels ---\n");
		fprintf(fw,"\tfunction pixels (col: std_logic_vector; row: std_logic_vector) return std_logic_vector is\n");
		fprintf(fw,"\t\tvariable pix: std_logic;\n");
		fprintf(fw,"\tbegin\n");

		// convert image
		int i,j;
		for(j=0;j<y;j++)
		{
			uchar *pix=(uchar*)(bmp->ScanLine[j]);
			for(i=0;i<x;i++)
			{
				fprintf(fw,"\t\t%s (col=%-3d and row=%-3d) then pix := '%d';\n",(i!=0 || j!=0)?"elsif":"   if",i,j,*pix?0:1);
				pix+=3;
			}
		}

		// generate rest of VHDL trash
		fprintf(fw,"\t\telse pix := '0'; end if;\n\n");
		fprintf(fw,"\t\treturn pix;\n");
		fprintf(fw,"\tend pixels;\n");
	}
	fprintf(fw,"end %s;\n",fname);

	// close destination file
	fclose(fw);

	// loose bitmap
	delete(bmp);

	// done
	printf("\ndone ...\n");

	return(0);
}
//---------------------------------------------------------------------------
